@@ -2,7 +2,7 @@
 
   etc.c -
 
-  $Author: nobu $
+  $Author$
   created at: Tue Mar 22 18:39:19 JST 1994
 
 ************************************************/
@@ -30,6 +30,10 @@
 #include <sys/utsname.h>
 #endif
 
+#ifdef HAVE_SCHED_GETAFFINITY
+#include <sched.h>
+#endif
+
 static VALUE sPasswd;
 #ifdef HAVE_GETGRENT
 static VALUE sGroup;
@@ -48,8 +52,33 @@ char *getenv();
 #endif
 char *getlogin();
 
+#define RUBY_ETC_VERSION "1.4.0"
+
+#ifdef HAVE_RB_DEPRECATE_CONSTANT
+void rb_deprecate_constant(VALUE mod, const char *name);
+#else
+# define rb_deprecate_constant(mod,name) ((void)(mod),(void)(name))
+#endif
+
 #include "constdefs.h"
 
+#ifdef HAVE_RUBY_ATOMIC_H
+# include "ruby/atomic.h"
+#else
+typedef int rb_atomic_t;
+# define RUBY_ATOMIC_CAS(var, oldval, newval) \
+    ((var) == (oldval) ? ((var) = (newval), (oldval)) : (var))
+# define RUBY_ATOMIC_EXCHANGE(var, newval) \
+    atomic_exchange(&var, newval)
+static inline rb_atomic_t
+atomic_exchange(volatile rb_atomic_t *var, rb_atomic_t newval)
+{
+    rb_atomic_t oldval = *var;
+    *var = newval;
+    return oldval;
+}
+#endif
+
 /* call-seq:
  *	getlogin	->  String
  *
@@ -77,8 +106,15 @@ etc_getlogin(VALUE obj)
     login = getenv("USER");
 #endif
 
-    if (login)
-	return rb_tainted_str_new2(login);
+    if (login) {
+#ifdef _WIN32
+	rb_encoding *extenc = rb_utf8_encoding();
+#else
+	rb_encoding *extenc = rb_locale_encoding();
+#endif
+	return rb_external_str_new_with_enc(login, strlen(login), extenc);
+    }
+
     return Qnil;
 }
 
@@ -87,7 +123,7 @@ static VALUE
 safe_setup_str(const char *str)
 {
     if (str == 0) str = "";
-    return rb_tainted_str_new2(str);
+    return rb_str_new2(str);
 }
 
 static VALUE
@@ -106,6 +142,12 @@ safe_setup_filesystem_str(const char *str)
 #endif
 
 #ifdef HAVE_GETPWENT
+# ifdef __APPLE__
+#   define PW_TIME2VAL(t) INT2NUM((int)(t))
+# else
+#   define PW_TIME2VAL(t) TIMET2NUM(t)
+# endif
+
 static VALUE
 setup_passwd(struct passwd *pwd)
 {
@@ -123,7 +165,7 @@ setup_passwd(struct passwd *pwd)
 			 safe_setup_filesystem_str(pwd->pw_dir),
 			 safe_setup_filesystem_str(pwd->pw_shell),
 #ifdef HAVE_STRUCT_PASSWD_PW_CHANGE
-			 INT2NUM(pwd->pw_change),
+			 PW_TIME2VAL(pwd->pw_change),
 #endif
 #ifdef HAVE_STRUCT_PASSWD_PW_QUOTA
 			 INT2NUM(pwd->pw_quota),
@@ -138,7 +180,7 @@ setup_passwd(struct passwd *pwd)
 			 safe_setup_locale_str(pwd->pw_comment),
 #endif
 #ifdef HAVE_STRUCT_PASSWD_PW_EXPIRE
-			 INT2NUM(pwd->pw_expire),
+			 PW_TIME2VAL(pwd->pw_expire),
 #endif
 			 0		/*dummy*/
 	);
@@ -204,9 +246,9 @@ etc_getpwnam(VALUE obj, VALUE nam)
 {
 #ifdef HAVE_GETPWENT
     struct passwd *pwd;
+    const char *p = StringValueCStr(nam);
 
-    SafeStringValue(nam);
-    pwd = getpwnam(RSTRING_PTR(nam));
+    pwd = getpwnam(p);
     if (pwd == 0) rb_raise(rb_eArgError, "can't find user for %"PRIsVALUE, nam);
     return setup_passwd(pwd);
 #else
@@ -215,22 +257,24 @@ etc_getpwnam(VALUE obj, VALUE nam)
 }
 
 #ifdef HAVE_GETPWENT
-static int passwd_blocking = 0;
+static rb_atomic_t passwd_blocking;
 static VALUE
-passwd_ensure(void)
+passwd_ensure(VALUE _)
 {
     endpwent();
-    passwd_blocking = (int)Qfalse;
+    if (RUBY_ATOMIC_EXCHANGE(passwd_blocking, 0) != 1) {
+	rb_raise(rb_eRuntimeError, "unexpected passwd_blocking");
+    }
     return Qnil;
 }
 
 static VALUE
-passwd_iterate(void)
+passwd_iterate(VALUE _)
 {
     struct passwd *pw;
 
     setpwent();
-    while (pw = getpwent()) {
+    while ((pw = getpwent()) != 0) {
 	rb_yield(setup_passwd(pw));
     }
     return Qnil;
@@ -239,10 +283,9 @@ passwd_iterate(void)
 static void
 each_passwd(void)
 {
-    if (passwd_blocking) {
+    if (RUBY_ATOMIC_CAS(passwd_blocking, 0, 1)) {
 	rb_raise(rb_eRuntimeError, "parallel passwd iteration");
     }
-    passwd_blocking = (int)Qtrue;
     rb_ensure(passwd_iterate, 0, passwd_ensure, 0);
 }
 #endif
@@ -276,7 +319,7 @@ etc_passwd(VALUE obj)
     if (rb_block_given_p()) {
 	each_passwd();
     }
-    else if (pw = getpwent()) {
+    else if ((pw = getpwent()) != 0) {
 	return setup_passwd(pw);
     }
 #endif
@@ -358,7 +401,7 @@ etc_getpwent(VALUE obj)
 #ifdef HAVE_GETPWENT
     struct passwd *pw;
 
-    if (pw = getpwent()) {
+    if ((pw = getpwent()) != 0) {
 	return setup_passwd(pw);
     }
 #endif
@@ -447,9 +490,9 @@ etc_getgrnam(VALUE obj, VALUE nam)
 {
 #ifdef HAVE_GETGRENT
     struct group *grp;
+    const char *p = StringValueCStr(nam);
 
-    SafeStringValue(nam);
-    grp = getgrnam(RSTRING_PTR(nam));
+    grp = getgrnam(p);
     if (grp == 0) rb_raise(rb_eArgError, "can't find group for %"PRIsVALUE, nam);
     return setup_group(grp);
 #else
@@ -458,23 +501,25 @@ etc_getgrnam(VALUE obj, VALUE nam)
 }
 
 #ifdef HAVE_GETGRENT
-static int group_blocking = 0;
+static rb_atomic_t group_blocking;
 static VALUE
-group_ensure(void)
+group_ensure(VALUE _)
 {
     endgrent();
-    group_blocking = (int)Qfalse;
+    if (RUBY_ATOMIC_EXCHANGE(group_blocking, 0) != 1) {
+	rb_raise(rb_eRuntimeError, "unexpected group_blocking");
+    }
     return Qnil;
 }
 
 
 static VALUE
-group_iterate(void)
+group_iterate(VALUE _)
 {
     struct group *pw;
 
     setgrent();
-    while (pw = getgrent()) {
+    while ((pw = getgrent()) != 0) {
 	rb_yield(setup_group(pw));
     }
     return Qnil;
@@ -483,10 +528,9 @@ group_iterate(void)
 static void
 each_group(void)
 {
-    if (group_blocking) {
+    if (RUBY_ATOMIC_CAS(group_blocking, 0, 1)) {
 	rb_raise(rb_eRuntimeError, "parallel group iteration");
     }
-    group_blocking = (int)Qtrue;
     rb_ensure(group_iterate, 0, group_ensure, 0);
 }
 #endif
@@ -516,7 +560,7 @@ etc_group(VALUE obj)
     if (rb_block_given_p()) {
 	each_group();
     }
-    else if (grp = getgrent()) {
+    else if ((grp = getgrent()) != 0) {
 	return setup_group(grp);
     }
 #endif
@@ -595,7 +639,7 @@ etc_getgrent(VALUE obj)
 #ifdef HAVE_GETGRENT
     struct group *gr;
 
-    if (gr = getgrent()) {
+    if ((gr = getgrent()) != 0) {
 	return setup_group(gr);
     }
 #endif
@@ -614,8 +658,9 @@ VALUE rb_w32_conv_from_wchar(const WCHAR *wstr, rb_encoding *enc);
  * Returns system configuration directory.
  *
  * This is typically "/etc", but is modified by the prefix used when Ruby was
- * compiled. For example, if Ruby is built and installed in /usr/local, returns
- * "/usr/local/etc".
+ * compiled. For example, if Ruby is built and installed in /usr/local,
+ * returns "/usr/local/etc" on other platforms than Windows.
+ * On Windows, this always returns the directory provided by the system.
  */
 static VALUE
 etc_sysconfdir(VALUE obj)
@@ -631,7 +676,7 @@ etc_sysconfdir(VALUE obj)
  * Returns system temporary directory; typically "/tmp".
  */
 static VALUE
-etc_systmpdir(void)
+etc_systmpdir(VALUE _)
 {
     VALUE tmpdir;
 #ifdef _WIN32
@@ -640,9 +685,33 @@ etc_systmpdir(void)
     if (!len) return Qnil;
     tmpdir = rb_w32_conv_from_wchar(path, rb_filesystem_encoding());
 #else
-    tmpdir = rb_filesystem_str_new_cstr("/tmp");
+    const char default_tmp[] = "/tmp";
+    const char *tmpstr = default_tmp;
+    size_t tmplen = strlen(default_tmp);
+# if defined _CS_DARWIN_USER_TEMP_DIR
+    #ifndef MAXPATHLEN
+    #define MAXPATHLEN 1024
+    #endif
+    char path[MAXPATHLEN];
+    size_t len;
+    len = confstr(_CS_DARWIN_USER_TEMP_DIR, path, sizeof(path));
+    if (len > 0) {
+	tmpstr = path;
+	tmplen = len - 1;
+	if (len > sizeof(path)) tmpstr = 0;
+    }
+# endif
+    tmpdir = rb_filesystem_str_new(tmpstr, tmplen);
+# if defined _CS_DARWIN_USER_TEMP_DIR
+    if (!tmpstr) {
+	confstr(_CS_DARWIN_USER_TEMP_DIR, RSTRING_PTR(tmpdir), len);
+    }
+# endif
 #endif
+#ifndef RB_PASS_KEYWORDS
+    /* untaint on Ruby < 2.7 */
     FL_UNSET(tmpdir, FL_TAINT);
+#endif
     return tmpdir;
 }
 
@@ -719,9 +788,6 @@ etc_uname(VALUE obj)
 # ifndef PROCESSOR_ARCHITECTURE_AMD64
 #   define PROCESSOR_ARCHITECTURE_AMD64 9
 # endif
-# ifndef PROCESSOR_ARCHITECTURE_IA64
-#   define PROCESSOR_ARCHITECTURE_IA64 6
-# endif
 # ifndef PROCESSOR_ARCHITECTURE_INTEL
 #   define PROCESSOR_ARCHITECTURE_INTEL 0
 # endif
@@ -733,9 +799,6 @@ etc_uname(VALUE obj)
       case PROCESSOR_ARCHITECTURE_ARM:
 	mach = "ARM";
 	break;
-      case PROCESSOR_ARCHITECTURE_IA64:
-	mach = "IA64";
-	break;
       case PROCESSOR_ARCHITECTURE_INTEL:
 	mach = "x86";
 	break;
@@ -778,10 +841,7 @@ etc_uname(VALUE obj)
  * nil means indefinite limit.  (sysconf() returns -1 but errno is not set.)
  *
  *   Etc.sysconf(Etc::SC_ARG_MAX) #=> 2097152
- *
- *   # Number of processors.
- *   # It is not standardized.
- *   Etc.sysconf(Etc::SC_NPROCESSORS_ONLN) #=> 4
+ *   Etc.sysconf(Etc::SC_LOGIN_NAME_MAX) #=> 256
  *
  */
 static VALUE
@@ -891,6 +951,120 @@ io_pathconf(VALUE io, VALUE arg)
 #define io_pathconf rb_f_notimplement
 #endif
 
+#if (defined(HAVE_SYSCONF) && defined(_SC_NPROCESSORS_ONLN)) || defined(_WIN32)
+
+#if defined(HAVE_SCHED_GETAFFINITY) && defined(CPU_ALLOC)
+static int
+etc_nprocessors_affin(void)
+{
+    cpu_set_t *cpuset, cpuset_buff[1024 / sizeof(cpu_set_t)];
+    size_t size;
+    int ret;
+    int n;
+
+    CPU_ZERO_S(sizeof(cpuset_buff), cpuset_buff);
+
+    /*
+     * XXX:
+     * man page says CPU_ALLOC takes number of cpus. But it is not accurate
+     * explanation. sched_getaffinity() returns EINVAL if cpuset bitmap is
+     * smaller than kernel internal bitmap.
+     * That said, sched_getaffinity() can fail when a kernel have sparse bitmap
+     * even if cpuset bitmap is larger than number of cpus.
+     * The precious way is to use /sys/devices/system/cpu/online. But there are
+     * two problems,
+     * - Costly calculation
+     *    It is a minor issue, but possibly kill a benefit of a parallel processing.
+     * - No guarantee to exist /sys/devices/system/cpu/online
+     *    This is an issue especially when using Linux containers.
+     * So, we use hardcode number for a workaround. Current linux kernel
+     * (Linux 3.17) support 8192 cpus at maximum. Then 16384 must be enough.
+     */
+    for (n=64; n <= 16384; n *= 2) {
+	size = CPU_ALLOC_SIZE(n);
+	if (size >= sizeof(cpuset_buff)) {
+	    cpuset = xcalloc(1, size);
+	    if (!cpuset)
+		return -1;
+	} else {
+	    cpuset = cpuset_buff;
+	}
+
+	ret = sched_getaffinity(0, size, cpuset);
+	if (ret == 0) {
+	    /* On success, count number of cpus. */
+	    ret = CPU_COUNT_S(size, cpuset);
+	}
+
+	if (size >= sizeof(cpuset_buff)) {
+	    xfree(cpuset);
+	}
+	if (ret > 0 || errno != EINVAL) {
+	    return ret;
+	}
+    }
+
+    return ret;
+}
+#endif
+
+/*
+ * Returns the number of online processors.
+ *
+ * The result is intended as the number of processes to
+ * use all available processors.
+ *
+ * This method is implemented using:
+ * - sched_getaffinity(): Linux
+ * - sysconf(_SC_NPROCESSORS_ONLN): GNU/Linux, NetBSD, FreeBSD, OpenBSD, DragonFly BSD, OpenIndiana, Mac OS X, AIX
+ *
+ * Example:
+ *
+ *   require 'etc'
+ *   p Etc.nprocessors #=> 4
+ *
+ * The result might be smaller number than physical cpus especially when ruby
+ * process is bound to specific cpus. This is intended for getting better
+ * parallel processing.
+ *
+ * Example: (Linux)
+ *
+ *   linux$ taskset 0x3 ./ruby -retc -e "p Etc.nprocessors"  #=> 2
+ *
+ */
+static VALUE
+etc_nprocessors(VALUE obj)
+{
+    long ret;
+
+#if !defined(_WIN32)
+
+#if defined(HAVE_SCHED_GETAFFINITY) && defined(CPU_ALLOC)
+    int ncpus;
+
+    ncpus = etc_nprocessors_affin();
+    if (ncpus != -1) {
+	return INT2NUM(ncpus);
+    }
+    /* fallback to _SC_NPROCESSORS_ONLN */
+#endif
+
+    errno = 0;
+    ret = sysconf(_SC_NPROCESSORS_ONLN);
+    if (ret == -1) {
+        rb_sys_fail("sysconf(_SC_NPROCESSORS_ONLN)");
+    }
+#else
+    SYSTEM_INFO si;
+    GetSystemInfo(&si);
+    ret = (long)si.dwNumberOfProcessors;
+#endif
+    return LONG2NUM(ret);
+}
+#else
+#define etc_nprocessors rb_f_notimplement
+#endif
+
 /*
  * The Etc module provides access to information typically stored in
  * files in the /etc directory on Unix systems.
@@ -922,7 +1096,11 @@ Init_etc(void)
 {
     VALUE mEtc;
 
+    #ifdef HAVE_RB_EXT_RACTOR_SAFE
+	RB_EXT_RACTOR_SAFE(true);
+    #endif
     mEtc = rb_define_module("Etc");
+    rb_define_const(mEtc, "VERSION", rb_str_new_cstr(RUBY_ETC_VERSION));
     init_constants(mEtc);
 
     rb_define_module_function(mEtc, "getlogin", etc_getlogin, 0);
@@ -946,6 +1124,7 @@ Init_etc(void)
     rb_define_module_function(mEtc, "sysconf", etc_sysconf, 1);
     rb_define_module_function(mEtc, "confstr", etc_confstr, 1);
     rb_define_method(rb_cIO, "pathconf", io_pathconf, 1);
+    rb_define_module_function(mEtc, "nprocessors", etc_nprocessors, 0);
 
     sPasswd =  rb_struct_define_under(mEtc, "Passwd",
 				      "name",
@@ -979,8 +1158,7 @@ Init_etc(void)
 #endif
 				      NULL);
 #if 0
-    /* Define-const: Passwd
-     *
+    /*
      * Passwd is a Struct that contains the following members:
      *
      * name::
@@ -1018,9 +1196,8 @@ Init_etc(void)
      * expire::
      *	    account expiration time(integer) must be compiled with +HAVE_STRUCT_PASSWD_PW_EXPIRE+
      */
-    rb_define_const(mEtc, "Passwd", sPasswd);
+    sPasswd = rb_define_class_under(mEtc, "Passwd", rb_cStruct);
 #endif
-    rb_define_const(rb_cStruct, "Passwd", sPasswd); /* deprecated name */
     rb_extend_object(sPasswd, rb_mEnumerable);
     rb_define_singleton_method(sPasswd, "each", etc_each_passwd, 0);
 
@@ -1032,8 +1209,7 @@ Init_etc(void)
 				    "gid", "mem", NULL);
 
 #if 0
-    /* Define-const: Group
-     *
+    /*
      * Group is a Struct that is only available when compiled with +HAVE_GETGRENT+.
      *
      * The struct contains the following members:
@@ -1041,11 +1217,10 @@ Init_etc(void)
      * name::
      *	    contains the name of the group as a String.
      * passwd::
-     *	    contains the encrypted password as a String. An 'x' is
+     *	    contains the encrypted password as a String. An <code>'x'</code> is
      *	    returned if password access to the group is not available; an empty
      *	    string is returned if no password is needed to obtain membership of
      *	    the group.
-     *
      *	    Must be compiled with +HAVE_STRUCT_GROUP_GR_PASSWD+.
      * gid::
      *	    contains the group's numeric ID as an integer.
@@ -1053,9 +1228,8 @@ Init_etc(void)
      *	    is an Array of Strings containing the short login names of the
      *	    members of the group.
      */
-    rb_define_const(mEtc, "Group", sGroup);
+    sGroup = rb_define_class_under(mEtc, "Group", rb_cStruct);
 #endif
-    rb_define_const(rb_cStruct, "Group", sGroup); /* deprecated name */
     rb_extend_object(sGroup, rb_mEnumerable);
     rb_define_singleton_method(sGroup, "each", etc_each_group, 0);
 #endif
