@@ -22,6 +22,7 @@
 
 require 'net/protocol'
 require 'uri'
+require 'resolv'
 autoload :OpenSSL, 'openssl'
 
 module Net   #:nodoc:
@@ -132,7 +133,7 @@ module Net   #:nodoc:
   #   puts res.class.name # => 'HTTPOK'
   #
   #   # Body
-  #   puts res.body if res.response_body_permitted?
+  #   puts res.body
   #
   # === Following Redirection
   #
@@ -327,6 +328,8 @@ module Net   #:nodoc:
   # HTTPInformation::                    1xx
   #   HTTPContinue::                        100
   #   HTTPSwitchProtocol::                  101
+  #   HTTPProcessing::                      102
+  #   HTTPEarlyHints::                      103
   # HTTPSuccess::                        2xx
   #   HTTPOK::                              200
   #   HTTPCreated::                         201
@@ -336,6 +339,7 @@ module Net   #:nodoc:
   #   HTTPResetContent::                    205
   #   HTTPPartialContent::                  206
   #   HTTPMultiStatus::                     207
+  #   HTTPAlreadyReported::                 208
   #   HTTPIMUsed::                          226
   # HTTPRedirection::                    3xx
   #   HTTPMultipleChoices::                 300
@@ -345,6 +349,7 @@ module Net   #:nodoc:
   #   HTTPNotModified::                     304
   #   HTTPUseProxy::                        305
   #   HTTPTemporaryRedirect::               307
+  #   HTTPPermanentRedirect::               308
   # HTTPClientError::                    4xx
   #   HTTPBadRequest::                      400
   #   HTTPUnauthorized::                    401
@@ -364,6 +369,7 @@ module Net   #:nodoc:
   #   HTTPUnsupportedMediaType::            415
   #   HTTPRequestedRangeNotSatisfiable::    416
   #   HTTPExpectationFailed::               417
+  #   HTTPMisdirectedRequest::              421
   #   HTTPUnprocessableEntity::             422
   #   HTTPLocked::                          423
   #   HTTPFailedDependency::                424
@@ -379,7 +385,10 @@ module Net   #:nodoc:
   #   HTTPServiceUnavailable::              503
   #   HTTPGatewayTimeOut::                  504
   #   HTTPVersionNotSupported::             505
+  #   HTTPVariantAlsoNegotiates::           506
   #   HTTPInsufficientStorage::             507
+  #   HTTPLoopDetected::                    508
+  #   HTTPNotExtended::                     510
   #   HTTPNetworkAuthenticationRequired::   511
   #
   # There is also the Net::HTTPBadResponse exception which is raised when
@@ -388,12 +397,11 @@ module Net   #:nodoc:
   class HTTP < Protocol
 
     # :stopdoc:
-    VERSION = "0.1.0"
+    VERSION = "0.2.0"
     Revision = %q$Revision$.split[1]
     HTTPVersion = '1.1'
     begin
       require 'zlib'
-      require 'stringio'  #for our purposes (unpacking gzip) lump these together
       HAVE_ZLIB=true
     rescue LoadError
       HAVE_ZLIB=false
@@ -524,14 +532,13 @@ module Net   #:nodoc:
     #
     #   { "cmd" => "search", "q" => "ruby", "max" => "50" }
     #
-    # This method also does Basic Authentication iff +url+.user exists.
+    # This method also does Basic Authentication if and only if +url+.user exists.
     # But userinfo for authentication is deprecated (RFC3986).
     # So this feature will be removed.
     #
     # Example:
     #
     #   require 'net/http'
-    #   require 'uri'
     #
     #   Net::HTTP.post_form URI('http://www.example.com/search.cgi'),
     #                       { "q" => "ruby", "max" => "50" }
@@ -973,6 +980,12 @@ module Net   #:nodoc:
     private :do_start
 
     def connect
+      if use_ssl?
+        # reference early to load OpenSSL before connecting,
+        # as OpenSSL may take time to load.
+        @ssl_context = OpenSSL::SSL::SSLContext.new
+      end
+
       if proxy? then
         conn_addr = proxy_address
         conn_port = proxy_port
@@ -981,17 +994,16 @@ module Net   #:nodoc:
         conn_port = port
       end
 
-      D "opening connection to #{conn_addr}:#{conn_port}..."
-      s = Timeout.timeout(@open_timeout, Net::OpenTimeout) {
-        begin
-          TCPSocket.open(conn_addr, conn_port, @local_host, @local_port)
-        rescue => e
-          raise e, "Failed to open TCP connection to " +
-            "#{conn_addr}:#{conn_port} (#{e.message})"
-        end
-      }
+      debug "opening connection to #{conn_addr}:#{conn_port}..."
+      begin
+        s = Socket.tcp conn_addr, conn_port, @local_host, @local_port, connect_timeout: @open_timeout
+      rescue => e
+        e = Net::OpenTimeout.new(e) if e.is_a?(Errno::ETIMEDOUT) #for compatibility with previous versions
+        raise e, "Failed to open TCP connection to " +
+          "#{conn_addr}:#{conn_port} (#{e.message})"
+      end
       s.setsockopt(Socket::IPPROTO_TCP, Socket::TCP_NODELAY, 1)
-      D "opened"
+      debug "opened"
       if use_ssl?
         if proxy?
           plain_sock = BufferedIO.new(s, read_timeout: @read_timeout,
@@ -1020,35 +1032,52 @@ module Net   #:nodoc:
             end
           end
         end
-        @ssl_context = OpenSSL::SSL::SSLContext.new
         @ssl_context.set_params(ssl_parameters)
         @ssl_context.session_cache_mode =
           OpenSSL::SSL::SSLContext::SESSION_CACHE_CLIENT |
           OpenSSL::SSL::SSLContext::SESSION_CACHE_NO_INTERNAL_STORE
         @ssl_context.session_new_cb = proc {|sock, sess| @ssl_session = sess }
-        D "starting SSL for #{conn_addr}:#{conn_port}..."
+
+        # Still do the post_connection_check below even if connecting
+        # to IP address
+        verify_hostname = @ssl_context.verify_hostname
+
+        # Server Name Indication (SNI) RFC 3546/6066
+        case @address
+        when Resolv::IPv4::Regex, Resolv::IPv6::Regex
+          # don't set SNI, as IP addresses in SNI is not valid
+          # per RFC 6066, section 3.
+
+          # Avoid openssl warning
+          @ssl_context.verify_hostname = false
+        else
+          ssl_host_address = @address
+        end
+
+        debug "starting SSL for #{conn_addr}:#{conn_port}..."
         s = OpenSSL::SSL::SSLSocket.new(s, @ssl_context)
         s.sync_close = true
-        # Server Name Indication (SNI) RFC 3546
-        s.hostname = @address if s.respond_to? :hostname=
+        s.hostname = ssl_host_address if s.respond_to?(:hostname=) && ssl_host_address
+
         if @ssl_session and
            Process.clock_gettime(Process::CLOCK_REALTIME) < @ssl_session.time.to_f + @ssl_session.timeout
           s.session = @ssl_session
         end
         ssl_socket_connect(s, @open_timeout)
-        if (@ssl_context.verify_mode != OpenSSL::SSL::VERIFY_NONE) && @ssl_context.verify_hostname
+        if (@ssl_context.verify_mode != OpenSSL::SSL::VERIFY_NONE) && verify_hostname
           s.post_connection_check(@address)
         end
-        D "SSL established, protocol: #{s.ssl_version}, cipher: #{s.cipher[0]}"
+        debug "SSL established, protocol: #{s.ssl_version}, cipher: #{s.cipher[0]}"
       end
       @socket = BufferedIO.new(s, read_timeout: @read_timeout,
                                write_timeout: @write_timeout,
                                continue_timeout: @continue_timeout,
                                debug_output: @debug_output)
+      @last_communicated = nil
       on_connect
     rescue => exception
       if s
-        D "Conn close because of connect error #{exception}"
+        debug "Conn close because of connect error #{exception}"
         s.close
       end
       raise
@@ -1180,7 +1209,8 @@ module Net   #:nodoc:
     # The username of the proxy server, if one is configured.
     def proxy_user
       if ENVIRONMENT_VARIABLE_IS_MULTIUSER_SAFE && @proxy_from_env
-        proxy_uri&.user
+        user = proxy_uri&.user
+        unescape(user) if user
       else
         @proxy_user
       end
@@ -1189,7 +1219,8 @@ module Net   #:nodoc:
     # The password of the proxy server, if one is configured.
     def proxy_pass
       if ENVIRONMENT_VARIABLE_IS_MULTIUSER_SAFE && @proxy_from_env
-        proxy_uri&.password
+        pass = proxy_uri&.password
+        unescape(pass) if pass
       else
         @proxy_pass
       end
@@ -1200,6 +1231,11 @@ module Net   #:nodoc:
 
     private
 
+    def unescape(value)
+      require 'cgi/util'
+      CGI.unescape(value)
+    end
+
     # without proxy, obsolete
 
     def conn_address # :nodoc:
@@ -1575,10 +1611,10 @@ module Net   #:nodoc:
         if count < max_retries && IDEMPOTENT_METHODS_.include?(req.method)
           count += 1
           @socket.close if @socket
-          D "Conn close because of error #{exception}, and retry"
+          debug "Conn close because of error #{exception}, and retry"
           retry
         end
-        D "Conn close because of error #{exception}"
+        debug "Conn close because of error #{exception}"
         @socket.close if @socket
         raise
       end
@@ -1586,7 +1622,7 @@ module Net   #:nodoc:
       end_transport req, res
       res
     rescue => exception
-      D "Conn close because of error #{exception}"
+      debug "Conn close because of error #{exception}"
       @socket.close if @socket
       raise exception
     end
@@ -1596,11 +1632,11 @@ module Net   #:nodoc:
         connect
       elsif @last_communicated
         if @last_communicated + @keep_alive_timeout < Process.clock_gettime(Process::CLOCK_MONOTONIC)
-          D 'Conn close because of keep_alive_timeout'
+          debug 'Conn close because of keep_alive_timeout'
           @socket.close
           connect
         elsif @socket.io.to_io.wait_readable(0) && @socket.eof?
-          D "Conn close because of EOF"
+          debug "Conn close because of EOF"
           @socket.close
           connect
         end
@@ -1618,15 +1654,15 @@ module Net   #:nodoc:
       @curr_http_version = res.http_version
       @last_communicated = nil
       if @socket.closed?
-        D 'Conn socket closed'
+        debug 'Conn socket closed'
       elsif not res.body and @close_on_empty_response
-        D 'Conn close'
+        debug 'Conn close'
         @socket.close
       elsif keep_alive?(req, res)
-        D 'Conn keep-alive'
+        debug 'Conn keep-alive'
         @last_communicated = Process.clock_gettime(Process::CLOCK_MONOTONIC)
       else
-        D 'Conn close'
+        debug 'Conn close'
         @socket.close
       end
     end
@@ -1681,11 +1717,14 @@ module Net   #:nodoc:
       default_port == port ? addr : "#{addr}:#{port}"
     end
 
-    def D(msg)
+    # Adds a message to debugging output
+    def debug(msg)
       return unless @debug_output
       @debug_output << msg
       @debug_output << "\n"
     end
+
+    alias_method :D, :debug
   end
 
 end
