@@ -28,9 +28,6 @@ enum {
     AREF_HASH_THRESHOLD = 10
 };
 
-const rb_iseq_t *rb_method_for_self_aref(VALUE name, VALUE arg, const struct rb_builtin_function *func);
-const rb_iseq_t *rb_method_for_self_aset(VALUE name, VALUE arg, const struct rb_builtin_function *func);
-
 VALUE rb_cStruct;
 static ID id_members, id_back_members, id_keyword_init;
 
@@ -190,6 +187,17 @@ struct_member_pos(VALUE s, VALUE name)
     }
 }
 
+/*
+ *  call-seq:
+ *    StructClass::members -> array_of_symbols
+ *
+ *  Returns the member names of the Struct descendant as an array:
+ *
+ *     Customer = Struct.new(:name, :address, :zip)
+ *     Customer.members # => [:name, :address, :zip]
+ *
+ */
+
 static VALUE
 rb_struct_s_members_m(VALUE klass)
 {
@@ -229,32 +237,6 @@ rb_struct_getmember(VALUE obj, ID id)
     UNREACHABLE_RETURN(Qnil);
 }
 
-static VALUE rb_struct_ref0(VALUE obj) {return RSTRUCT_GET(obj, 0);}
-static VALUE rb_struct_ref1(VALUE obj) {return RSTRUCT_GET(obj, 1);}
-static VALUE rb_struct_ref2(VALUE obj) {return RSTRUCT_GET(obj, 2);}
-static VALUE rb_struct_ref3(VALUE obj) {return RSTRUCT_GET(obj, 3);}
-static VALUE rb_struct_ref4(VALUE obj) {return RSTRUCT_GET(obj, 4);}
-static VALUE rb_struct_ref5(VALUE obj) {return RSTRUCT_GET(obj, 5);}
-static VALUE rb_struct_ref6(VALUE obj) {return RSTRUCT_GET(obj, 6);}
-static VALUE rb_struct_ref7(VALUE obj) {return RSTRUCT_GET(obj, 7);}
-static VALUE rb_struct_ref8(VALUE obj) {return RSTRUCT_GET(obj, 8);}
-static VALUE rb_struct_ref9(VALUE obj) {return RSTRUCT_GET(obj, 9);}
-
-#define N_REF_FUNC numberof(ref_func)
-
-static VALUE (*const ref_func[])(VALUE) = {
-    rb_struct_ref0,
-    rb_struct_ref1,
-    rb_struct_ref2,
-    rb_struct_ref3,
-    rb_struct_ref4,
-    rb_struct_ref5,
-    rb_struct_ref6,
-    rb_struct_ref7,
-    rb_struct_ref8,
-    rb_struct_ref9,
-};
-
 static void
 rb_struct_modify(VALUE s)
 {
@@ -292,50 +274,16 @@ new_struct(VALUE name, VALUE super)
 
 NORETURN(static void invalid_struct_pos(VALUE s, VALUE idx));
 
-static inline long
-struct_pos_num(VALUE s, VALUE idx)
-{
-    long i = NUM2INT(idx);
-    if (i < 0 || i >= RSTRUCT_LEN(s)) invalid_struct_pos(s, idx);
-    return i;
-}
-
-static VALUE
-opt_struct_aref(rb_execution_context_t *ec, VALUE self, VALUE idx)
-{
-    long i = struct_pos_num(self, idx);
-    return RSTRUCT_GET(self, i);
-}
-
-static VALUE
-opt_struct_aset(rb_execution_context_t *ec, VALUE self, VALUE val, VALUE idx)
-{
-    long i = struct_pos_num(self, idx);
-    rb_struct_modify(self);
-    RSTRUCT_SET(self, i, val);
-    return val;
-}
-
-static const struct rb_builtin_function struct_aref_builtin =
-    RB_BUILTIN_FUNCTION(0, struct_aref, opt_struct_aref, 1, 0);
-static const struct rb_builtin_function struct_aset_builtin =
-    RB_BUILTIN_FUNCTION(1, struct_aref, opt_struct_aset, 2, 0);
-
 static void
 define_aref_method(VALUE nstr, VALUE name, VALUE off)
 {
-    const rb_iseq_t *iseq = rb_method_for_self_aref(name, off, &struct_aref_builtin);
-    iseq->body->builtin_inline_p = true;
-
-    rb_add_method_iseq(nstr, SYM2ID(name), iseq, NULL, METHOD_VISI_PUBLIC);
+    rb_add_method_optimized(nstr, SYM2ID(name), OPTIMIZED_METHOD_TYPE_STRUCT_AREF, FIX2UINT(off), METHOD_VISI_PUBLIC);
 }
 
 static void
 define_aset_method(VALUE nstr, VALUE name, VALUE off)
 {
-    const rb_iseq_t *iseq = rb_method_for_self_aset(name, off, &struct_aset_builtin);
-
-    rb_add_method_iseq(nstr, SYM2ID(name), iseq, NULL, METHOD_VISI_PUBLIC);
+    rb_add_method_optimized(nstr, SYM2ID(name), OPTIMIZED_METHOD_TYPE_STRUCT_ASET, FIX2UINT(off), METHOD_VISI_PUBLIC);
 }
 
 static VALUE
@@ -348,12 +296,14 @@ rb_struct_s_inspect(VALUE klass)
     return inspect;
 }
 
+#if 0 /* for RDoc */
+
 /*
  * call-seq:
- *   StructClass.keyword_init? -> true or false
+ *   StructClass::keyword_init? -> true or falsy value
  *
- * Returns true if the class was initialized with +keyword_init: true+.
- * Otherwise returns false.
+ * Returns +true+ if the class was initialized with <tt>keyword_init: true</tt>.
+ * Otherwise returns +nil+ or +false+.
  *
  * Examples:
  *   Foo = Struct.new(:a)
@@ -363,6 +313,11 @@ rb_struct_s_inspect(VALUE klass)
  *   Baz = Struct.new(:a, keyword_init: false)
  *   Baz.keyword_init? # => false
  */
+static VALUE
+rb_struct_s_keyword_init_p(VALUE obj)
+{
+}
+#endif
 
 #define rb_struct_s_keyword_init_p rb_struct_s_keyword_init
 
@@ -386,13 +341,8 @@ setup_struct(VALUE nstr, VALUE members)
         ID id = SYM2ID(sym);
 	VALUE off = LONG2NUM(i);
 
-	if (i < N_REF_FUNC) {
-	    rb_define_method_id(nstr, id, ref_func[i], 0);
-	}
-	else {
-            define_aref_method(nstr, sym, off);
-	}
-	define_aset_method(nstr, ID2SYM(rb_id_attrset(id)), off);
+        define_aref_method(nstr, sym, off);
+        define_aset_method(nstr, ID2SYM(rb_id_attrset(id)), off);
     }
 
     return nstr;
@@ -626,8 +576,9 @@ rb_struct_s_def(int argc, VALUE *argv, VALUE klass)
     long i;
     VALUE st;
     st_table *tbl;
+    VALUE opt;
 
-    rb_check_arity(argc, 1, UNLIMITED_ARGUMENTS);
+    argc = rb_scan_args(argc, argv, "1*:", NULL, NULL, &opt);
     name = argv[0];
     if (SYMBOL_P(name)) {
 	name = Qnil;
@@ -637,20 +588,19 @@ rb_struct_s_def(int argc, VALUE *argv, VALUE klass)
 	++argv;
     }
 
-    if (RB_TYPE_P(argv[argc-1], T_HASH)) {
+    if (!NIL_P(opt)) {
 	static ID keyword_ids[1];
 
 	if (!keyword_ids[0]) {
 	    keyword_ids[0] = rb_intern("keyword_init");
 	}
-        rb_get_kwargs(argv[argc-1], keyword_ids, 0, 1, &keyword_init);
+        rb_get_kwargs(opt, keyword_ids, 0, 1, &keyword_init);
         if (keyword_init == Qundef) {
             keyword_init = Qnil;
         }
         else if (RTEST(keyword_init)) {
             keyword_init = Qtrue;
         }
-	--argc;
     }
 
     rest = rb_ident_hash_new();
@@ -735,12 +685,25 @@ rb_struct_initialize_m(int argc, const VALUE *argv, VALUE self)
         return Qnil;
     }
 
-    VALUE keyword_init = rb_struct_s_keyword_init(klass);
-    if (RTEST(keyword_init)) {
-	struct struct_hash_set_arg arg;
+    bool keyword_init = false;
+    switch (rb_struct_s_keyword_init(klass)) {
+      default:
 	if (argc > 1 || !RB_TYPE_P(argv[0], T_HASH)) {
 	    rb_raise(rb_eArgError, "wrong number of arguments (given %d, expected 0)", argc);
 	}
+	keyword_init = true;
+	break;
+      case Qfalse:
+        break;
+      case Qnil:
+	if (argc > 1 || !RB_TYPE_P(argv[0], T_HASH)) {
+            break;
+        }
+	keyword_init = rb_keyword_given_p();
+        break;
+    }
+    if (keyword_init) {
+	struct struct_hash_set_arg arg;
 	rb_mem_clear((VALUE *)RSTRUCT_CONST_PTR(self), n);
 	arg.self = self;
 	arg.unknown_keywords = Qnil;
@@ -754,10 +717,6 @@ rb_struct_initialize_m(int argc, const VALUE *argv, VALUE self)
 	if (n < argc) {
 	    rb_raise(rb_eArgError, "struct size differs");
 	}
-        if (NIL_P(keyword_init) && argc == 1 && RB_TYPE_P(argv[0], T_HASH) && rb_keyword_given_p()) {
-            rb_warn("Passing only keyword arguments to Struct#initialize will behave differently from Ruby 3.2. "\
-                    "Please use a Hash literal like .new({k: v}) instead of .new(k: v).");
-        }
         for (long i=0; i<argc; i++) {
 	    RSTRUCT_SET(self, i, argv[i]);
 	}
@@ -844,7 +803,7 @@ rb_struct_alloc(VALUE klass, VALUE values)
 VALUE
 rb_struct_new(VALUE klass, ...)
 {
-    VALUE tmpargs[N_REF_FUNC], *mem = tmpargs;
+    VALUE tmpargs[16], *mem = tmpargs;
     int size, i;
     va_list args;
 
@@ -1210,7 +1169,7 @@ invalid_struct_pos(VALUE s, VALUE idx)
  *
  *  With integer argument +n+ given, returns <tt>self.values[n]</tt>
  *  if +n+ is in range;
- *  see {Array Indexes}[Array.html#class-Array-label-Array+Indexes]:
+ *  see Array@Array+Indexes:
  *
  *    joe[2]  # => 12345
  *    joe[-2] # => "123 Maple, Anytown NC"
@@ -1246,7 +1205,7 @@ rb_struct_aref(VALUE s, VALUE idx)
  *
  *  With integer argument +n+ given, assigns the given +value+
  *  to the +n+-th member if +n+ is in range;
- *  see {Array Indexes}[Array.html#class-Array-label-Array+Indexes]:
+ *  see Array@Array+Indexes:
  *
  *    joe = Customer.new("Joe Smith", "123 Maple, Anytown NC", 12345)
  *    joe[2] = 54321           # => 54321
@@ -1308,7 +1267,7 @@ struct_entry(VALUE s, long n)
  *    joe.values_at(0, -3)   # => ["Joe Smith", "Joe Smith"]
  *
  *  Raises IndexError if any of +integers+ is out of range;
- *  see {Array Indexes}[Array.html#class-Array-label-Array+Indexes].
+ *  see Array@Array+Indexes.
  *
  *  With integer range argument +integer_range+ given,
  *  returns an array containing each value given by the elements of the range;
@@ -1321,7 +1280,7 @@ struct_entry(VALUE s, long n)
  *    joe.values_at(1..4) # => ["123 Maple, Anytown NC", 12345, nil, nil]
  *
  *  Raises RangeError if any element of the range is negative and out of range;
- *  see {Array Indexes}[Array.html#class-Array-label-Array+Indexes].
+ *  see Array@Array+Indexes.
  *
  */
 
@@ -1603,62 +1562,62 @@ rb_struct_dig(int argc, VALUE *argv, VALUE self)
  *
  *  First, what's elsewhere. \Class \Struct:
  *
- *  - Inherits from {class Object}[Object.html#class-Object-label-What-27s+Here].
- *  - Includes {module Enumerable}[Enumerable.html#module-Enumerable-label-What-27s+Here],
+ *  - Inherits from {class Object}[rdoc-ref:Object@What-27s+Here].
+ *  - Includes {module Enumerable}[rdoc-ref:Enumerable@What-27s+Here],
  *    which provides dozens of additional methods.
  *
  *  Here, class \Struct provides methods that are useful for:
  *
- *  - {Creating a Struct Subclass}[#class-Struct-label-Methods+for+Creating+a+Struct+Subclass]
- *  - {Querying}[#class-Struct-label-Methods+for+Querying]
- *  - {Comparing}[#class-Struct-label-Methods+for+Comparing]
- *  - {Fetching}[#class-Struct-label-Methods+for+Fetching]
- *  - {Assigning}[#class-Struct-label-Methods+for+Assigning]
- *  - {Iterating}[#class-Struct-label-Methods+for+Iterating]
- *  - {Converting}[#class-Struct-label-Methods+for+Converting]
+ *  - {Creating a Struct Subclass}[rdoc-ref:Struct@Methods+for+Creating+a+Struct+Subclass]
+ *  - {Querying}[rdoc-ref:Struct@Methods+for+Querying]
+ *  - {Comparing}[rdoc-ref:Struct@Methods+for+Comparing]
+ *  - {Fetching}[rdoc-ref:Struct@Methods+for+Fetching]
+ *  - {Assigning}[rdoc-ref:Struct@Methods+for+Assigning]
+ *  - {Iterating}[rdoc-ref:Struct@Methods+for+Iterating]
+ *  - {Converting}[rdoc-ref:Struct@Methods+for+Converting]
  *
  *  === Methods for Creating a Struct Subclass
  *
- *  ::new:: Returns a new subclass of \Struct.
+ *  - ::new: Returns a new subclass of \Struct.
  *
  *  === Methods for Querying
  *
- *  #hash:: Returns the integer hash code.
- *  #length, #size:: Returns the number of members.
+ *  - #hash: Returns the integer hash code.
+ *  - #length, #size: Returns the number of members.
  *
  *  === Methods for Comparing
  *
- *  {#==}[#method-i-3D-3D]:: Returns whether a given object is equal to +self+,
- *                           using <tt>==</tt> to compare member values.
- *  #eql?:: Returns whether a given object is equal to +self+,
- *          using <tt>eql?</tt> to compare member values.
+ *  - #==: Returns whether a given object is equal to +self+, using <tt>==</tt>
+ *    to compare member values.
+ *  - #eql?: Returns whether a given object is equal to +self+,
+ *    using <tt>eql?</tt> to compare member values.
  *
  *  === Methods for Fetching
  *
- *  #[]:: Returns the value associated with a given member name.
- *  #to_a, #values, #deconstruct:: Returns the member values in +self+ as an array.
- *  #deconstruct_keys:: Returns a hash of the name/value pairs
- *                      for given member names.
- *  #dig:: Returns the object in nested objects that is specified
- *         by a given member name and additional arguments.
- *  #members:: Returns an array of the member names.
- *  #select, #filter:: Returns an array of member values from +self+,
- *                     as selected by the given block.
- *  #values_at:: Returns an array containing values for given member names.
+ *  - #[]: Returns the value associated with a given member name.
+ *  - #to_a, #values, #deconstruct: Returns the member values in +self+ as an array.
+ *  - #deconstruct_keys: Returns a hash of the name/value pairs
+ *    for given member names.
+ *  - #dig: Returns the object in nested objects that is specified
+ *    by a given member name and additional arguments.
+ *  - #members: Returns an array of the member names.
+ *  - #select, #filter: Returns an array of member values from +self+,
+ *    as selected by the given block.
+ *  - #values_at: Returns an array containing values for given member names.
  *
  *  === Methods for Assigning
  *
- *  #[]=:: Assigns a given value to a given member name.
+ *  - #[]=: Assigns a given value to a given member name.
  *
  *  === Methods for Iterating
  *
- *  #each:: Calls a given block with each member name.
- *  #each_pair:: Calls a given block with each member name/value pair.
+ *  - #each: Calls a given block with each member name.
+ *  - #each_pair: Calls a given block with each member name/value pair.
  *
  *  === Methods for Converting
  *
- *  #inspect, #to_s:: Returns a string representation of +self+.
- *  #to_h:: Returns a hash of the member name/value pairs in +self+.
+ *  - #inspect, #to_s: Returns a string representation of +self+.
+ *  - #to_h: Returns a hash of the member name/value pairs in +self+.
  *
  */
 void
@@ -1669,6 +1628,10 @@ InitVM_Struct(void)
 
     rb_undef_alloc_func(rb_cStruct);
     rb_define_singleton_method(rb_cStruct, "new", rb_struct_s_def, -1);
+#if 0 /* for RDoc */
+    rb_define_singleton_method(rb_cStruct, "keyword_init?", rb_struct_s_keyword_init_p, 0);
+    rb_define_singleton_method(rb_cStruct, "members", rb_struct_s_members_m, 0);
+#endif
 
     rb_define_method(rb_cStruct, "initialize", rb_struct_initialize_m, -1);
     rb_define_method(rb_cStruct, "initialize_copy", rb_struct_init_copy, 1);
