@@ -10,7 +10,6 @@ class TestBignum < Test::Unit::TestCase
   FIXNUM_MAX = RbConfig::LIMITS['FIXNUM_MAX']
 
   BIGNUM_MIN = FIXNUM_MAX + 1
-  b = BIGNUM_MIN
 
   f = BIGNUM_MIN
   n = 0
@@ -20,23 +19,22 @@ class TestBignum < Test::Unit::TestCase
   end
   BIGNUM_MIN_BITS = n
 
-  T_ZERO = 0.to_bignum
-  T_ONE  = 1.to_bignum
-  T_MONE = (-1).to_bignum
-  T31  = (2**31).to_bignum   # 2147483648
-  T31P = (T31 - 1).to_bignum # 2147483647
-  T32  = (2**32).to_bignum   # 4294967296
-  T32P = (T32 - 1).to_bignum # 4294967295
-  T64  = (2**64).to_bignum   # 18446744073709551616
-  T64P = (T64 - 1).to_bignum # 18446744073709551615
-  T128  = (2**128).to_bignum
-  T128P = (T128 - 1).to_bignum
-  T1024  = (2**1024).to_bignum
-  T1024P = (T1024 - 1).to_bignum
+  T_ZERO = Bug::Integer.to_bignum(0)
+  T_ONE  = Bug::Integer.to_bignum(1)
+  T_MONE = Bug::Integer.to_bignum(-1)
+  T31  = Bug::Integer.to_bignum(2**31)   # 2147483648
+  T31P = Bug::Integer.to_bignum(T31 - 1) # 2147483647
+  T32  = Bug::Integer.to_bignum(2**32)   # 4294967296
+  T32P = Bug::Integer.to_bignum(T32 - 1) # 4294967295
+  T64  = Bug::Integer.to_bignum(2**64)   # 18446744073709551616
+  T64P = Bug::Integer.to_bignum(T64 - 1) # 18446744073709551615
+  T128  = Bug::Integer.to_bignum(2**128)
+  T128P = Bug::Integer.to_bignum(T128 - 1)
+  T1024  = Bug::Integer.to_bignum(2**1024)
+  T1024P = Bug::Integer.to_bignum(T1024 - 1)
 
   def setup
     @verbose = $VERBOSE
-    $VERBOSE = nil
     @fmax = Float::MAX.to_i
     @fmax2 = @fmax * 2
     @big = (1 << BIGNUM_MIN_BITS) - 1
@@ -215,9 +213,11 @@ class TestBignum < Test::Unit::TestCase
 
   def test_to_f
     assert_nothing_raised { T31P.to_f.to_i }
-    assert_raise(FloatDomainError) { (1024**1024).to_f.to_i }
-    assert_equal(1, (2**50000).to_f.infinite?)
-    assert_equal(-1, (-(2**50000)).to_f.infinite?)
+    assert_raise(FloatDomainError) {
+      assert_warning(/out of Float range/) {(1024**1024).to_f}.to_i
+    }
+    assert_equal(1, assert_warning(/out of Float range/) {(2**50000).to_f}.infinite?)
+    assert_equal(-1, assert_warning(/out of Float range/) {(-(2**50000)).to_f}.infinite?)
   end
 
   def test_cmp
@@ -416,7 +416,7 @@ class TestBignum < Test::Unit::TestCase
   def test_divide
     bug5490 = '[ruby-core:40429]'
     assert_raise(ZeroDivisionError, bug5490) {T1024./(0)}
-    assert_equal(Float::INFINITY, T1024./(0.0), bug5490)
+    assert_equal(Float::INFINITY, assert_warning(/out of Float range/) {T1024./(0.0)}, bug5490)
   end
 
   def test_div
@@ -467,8 +467,8 @@ class TestBignum < Test::Unit::TestCase
   def test_pow
     assert_equal(1.0, T32 ** 0.0)
     assert_equal(1.0 / T32, T32 ** -1)
-    assert_equal(1, (T32 ** T32).infinite?)
-    assert_equal(1, (T32 ** (2**30-1)).infinite?)
+    assert_equal(1, assert_warning(/may be too big/) {T32 ** T32}.infinite?)
+    assert_equal(1, assert_warning(/may be too big/) {T32 ** (2**30-1)}.infinite?)
 
     ### rational changes the behavior of Bignum#**
     #assert_raise(TypeError) { T32**"foo" }
@@ -506,39 +506,57 @@ class TestBignum < Test::Unit::TestCase
   end
 
   def test_and_with_float
-    assert_raise(TypeError) { T1024 & 1.5 }
+    assert_raise(TypeError) {
+      assert_warning(/out of Float range/) {T1024 & 1.5}
+    }
   end
 
   def test_and_with_rational
-    assert_raise(TypeError, "#1792") { T1024 & Rational(3, 2) }
+    assert_raise(TypeError, "#1792") {
+      assert_warn(/out of Float range/) {T1024 & Rational(3, 2)}
+    }
   end
 
   def test_and_with_nonintegral_numeric
-    assert_raise(TypeError, "#1792") { T1024 & DummyNumeric.new }
+    assert_raise(TypeError, "#1792") {
+      assert_warn(/out of Float range/) {T1024 & DummyNumeric.new}
+    }
   end
 
   def test_or_with_float
-    assert_raise(TypeError) { T1024 | 1.5 }
+    assert_raise(TypeError) {
+      assert_warn(/out of Float range/) {T1024 | 1.5}
+    }
   end
 
   def test_or_with_rational
-    assert_raise(TypeError, "#1792") { T1024 | Rational(3, 2) }
+    assert_raise(TypeError, "#1792") {
+      assert_warn(/out of Float range/) {T1024 | Rational(3, 2)}
+    }
   end
 
   def test_or_with_nonintegral_numeric
-    assert_raise(TypeError, "#1792") { T1024 | DummyNumeric.new }
+    assert_raise(TypeError, "#1792") {
+      assert_warn(/out of Float range/) {T1024 | DummyNumeric.new}
+    }
   end
 
   def test_xor_with_float
-    assert_raise(TypeError) { T1024 ^ 1.5 }
+    assert_raise(TypeError) {
+      assert_warn(/out of Float range/) {T1024 ^ 1.5}
+    }
   end
 
   def test_xor_with_rational
-    assert_raise(TypeError, "#1792") { T1024 ^ Rational(3, 2) }
+    assert_raise(TypeError, "#1792") {
+      assert_warn(/out of Float range/) {T1024 ^ Rational(3, 2)}
+    }
   end
 
   def test_xor_with_nonintegral_numeric
-    assert_raise(TypeError, "#1792") { T1024 ^ DummyNumeric.new }
+    assert_raise(TypeError, "#1792") {
+      assert_warn(/out of Float range/) {T1024 ^ DummyNumeric.new}
+    }
   end
 
   def test_shift2
@@ -612,19 +630,21 @@ class TestBignum < Test::Unit::TestCase
       return # GMP doesn't support interrupt during an operation.
     end
     time = Time.now
-    start_flag = false
     end_flag = false
     num = (65536 ** 65536)
+    q = Thread::Queue.new
     thread = Thread.new do
-      start_flag = true
-      num.to_s
-      end_flag = true
+      assert_raise(RuntimeError) {
+        q << true
+        num.to_s
+        end_flag = true
+      }
     end
-    sleep 0.001 until start_flag
+    q.pop # sync
     thread.raise
-    thread.join rescue nil
+    thread.join
     time = Time.now - time
-    skip "too fast cpu" if end_flag
+    omit "too fast cpu" if end_flag
     assert_operator(time, :<, 10)
   end
 
@@ -655,14 +675,14 @@ class TestBignum < Test::Unit::TestCase
           return
         end
       end
-      skip "cannot create suitable test case"
+      omit "cannot create suitable test case"
     ensure
       Signal.trap(:INT, oldtrap) if oldtrap
     end
   end
 
   def test_too_big_to_s
-    if (big = 2**31-1).fixnum?
+    if Bug::Integer.fixnum?(big = 2**31-1)
       return
     end
     assert_raise_with_message(RangeError, /too big to convert/) {(1 << big).to_s}
@@ -745,7 +765,7 @@ class TestBignum < Test::Unit::TestCase
   end
 
   def test_digits
-    assert_equal([90, 78, 56, 34, 12], 1234567890.to_bignum.digits(100))
+    assert_equal([90, 78, 56, 34, 12], Bug::Integer.to_bignum(1234567890).digits(100))
     assert_equal([7215, 2413, 6242], T1024P.digits(10_000).first(3))
     assert_equal([11], 11.digits(T1024P))
     assert_equal([T1024P - 1, 1], (T1024P + T1024P - 1).digits(T1024P))
@@ -758,13 +778,13 @@ class TestBignum < Test::Unit::TestCase
   end
 
   def test_digits_for_invalid_base_numbers
-    assert_raise(ArgumentError) { T1024P.to_bignum.digits(0) }
-    assert_raise(ArgumentError) { T1024P.to_bignum.digits(-1) }
-    assert_raise(ArgumentError) { T1024P.to_bignum.digits(0.to_bignum) }
-    assert_raise(ArgumentError) { T1024P.to_bignum.digits(1.to_bignum) }
-    assert_raise(ArgumentError) { T1024P.to_bignum.digits(-T1024P) }
-    assert_raise(ArgumentError) { 10.digits(0.to_bignum) }
-    assert_raise(ArgumentError) { 10.digits(1.to_bignum) }
+    assert_raise(ArgumentError) { Bug::Integer.to_bignum(T1024P).digits(0) }
+    assert_raise(ArgumentError) { Bug::Integer.to_bignum(T1024P).digits(-1) }
+    assert_raise(ArgumentError) { Bug::Integer.to_bignum(T1024P).digits(Bug::Integer.to_bignum(0)) }
+    assert_raise(ArgumentError) { Bug::Integer.to_bignum(T1024P).digits(Bug::Integer.to_bignum(1)) }
+    assert_raise(ArgumentError) { Bug::Integer.to_bignum(T1024P).digits(-T1024P) }
+    assert_raise(ArgumentError) { 10.digits(Bug::Integer.to_bignum(0)) }
+    assert_raise(ArgumentError) { 10.digits(Bug::Integer.to_bignum(1)) }
   end
 
   def test_digits_for_non_integral_base_numbers
