@@ -64,7 +64,7 @@ class TestJIT < Test::Unit::TestCase
 
   def setup
     unless JITSupport.supported?
-      skip 'JIT seems not supported on this platform'
+      omit 'JIT seems not supported on this platform'
     end
     self.class.setup
   end
@@ -243,8 +243,8 @@ class TestJIT < Test::Unit::TestCase
     end;
   end
 
-  def test_compile_insn_putstring_concatstrings_tostring
-    assert_compile_once('"a#{}b" + "c"', result_inspect: '"abc"', insns: %i[putstring concatstrings tostring])
+  def test_compile_insn_putstring_concatstrings_objtostring
+    assert_compile_once('"a#{}b" + "c"', result_inspect: '"abc"', insns: %i[putstring concatstrings objtostring])
   end
 
   def test_compile_insn_toregexp
@@ -321,7 +321,7 @@ class TestJIT < Test::Unit::TestCase
   end
 
   def test_compile_insn_reput
-    skip "write test"
+    omit "write test"
   end
 
   def test_compile_insn_setn
@@ -352,11 +352,11 @@ class TestJIT < Test::Unit::TestCase
   end
 
   def test_compile_insn_tracecoverage
-    skip "write test"
+    omit "write test"
   end
 
   def test_compile_insn_defineclass
-    skip "support this in mjit_compile (low priority)"
+    omit "support this in mjit_compile (low priority)"
   end
 
   def test_compile_insn_send
@@ -482,8 +482,8 @@ class TestJIT < Test::Unit::TestCase
     end;
   end
 
-  def test_compile_insn_checktype
-    assert_compile_once("#{<<~"begin;"}\n#{<<~'end;'}", result_inspect: '"42"', insns: %i[checktype])
+  def test_compile_insn_objtostring
+    assert_compile_once("#{<<~"begin;"}\n#{<<~'end;'}", result_inspect: '"42"', insns: %i[objtostring])
     begin;
       a = '2'
       "4#{a}"
@@ -709,7 +709,7 @@ class TestJIT < Test::Unit::TestCase
 
       if RUBY_PLATFORM.match?(/mswin/)
         # "Permission Denied" error is preventing to remove so file on AppVeyor/RubyCI.
-        skip 'Removing so file is randomly failing on AppVeyor/RubyCI mswin due to Permission Denied.'
+        omit 'Removing so file is randomly failing on AppVeyor/RubyCI mswin due to Permission Denied.'
       else
         # verify .c files are deleted on unload_units
         assert_send([Dir, :empty?, dir], debug_info)
@@ -996,7 +996,7 @@ class TestJIT < Test::Unit::TestCase
 
   def test_clean_so
     if RUBY_PLATFORM.match?(/mswin/)
-      skip 'Removing so file is randomly failing on AppVeyor/RubyCI mswin due to Permission Denied.'
+      omit 'Removing so file is randomly failing on AppVeyor/RubyCI mswin due to Permission Denied.'
     end
     Dir.mktmpdir("jit_test_clean_so_") do |dir|
       code = "x = 0; 10.times {|i|x+=i}"
@@ -1010,7 +1010,7 @@ class TestJIT < Test::Unit::TestCase
   def test_clean_objects_on_exec
     if /mswin|mingw/ =~ RUBY_PLATFORM
       # TODO: check call stack and close handle of code which is not on stack, and remove objects on best-effort basis
-      skip 'Removing so file being used does not work on Windows'
+      omit 'Removing so file being used does not work on Windows'
     end
     Dir.mktmpdir("jit_test_clean_objects_on_exec_") do |dir|
       eval_with_jit({"TMPDIR"=>dir}, "#{<<~"begin;"}\n#{<<~"end;"}", min_calls: 1)
@@ -1108,7 +1108,7 @@ class TestJIT < Test::Unit::TestCase
   def test_mjit_pause_wait
     assert_eval_with_jit("#{<<~"begin;"}\n#{<<~"end;"}", stdout: '', success_count: 0, min_calls: 1)
     begin;
-      RubyVM::JIT.pause
+      RubyVM::MJIT.pause
       proc {}.call
     end;
   end
@@ -1194,8 +1194,8 @@ class TestJIT < Test::Unit::TestCase
     out, err = eval_with_jit(script, verbose: 1, min_calls: min_calls, max_cache: max_cache)
     success_actual = err.scan(/^#{JIT_SUCCESS_PREFIX}:/).size
     recompile_actual = err.scan(/^#{JIT_RECOMPILE_PREFIX}:/).size
-    # Add --jit-verbose=2 logs for cl.exe because compiler's error message is suppressed
-    # for cl.exe with --jit-verbose=1. See `start_process` in mjit_worker.c.
+    # Add --mjit-verbose=2 logs for cl.exe because compiler's error message is suppressed
+    # for cl.exe with --mjit-verbose=1. See `start_process` in mjit_worker.c.
     if RUBY_PLATFORM.match?(/mswin/) && success_count != success_actual
       out2, err2 = eval_with_jit(script, verbose: 2, min_calls: min_calls, max_cache: max_cache)
     end
@@ -1231,7 +1231,9 @@ class TestJIT < Test::Unit::TestCase
   end
 
   def mark_tested_insn(insn, used_insns:, uplevel: 1)
-    unless used_insns.include?(insn)
+    # Currently, this check emits a false-positive warning against opt_regexpmatch2,
+    # so the insn is excluded explicitly. See https://bugs.ruby-lang.org/issues/18269
+    if !used_insns.include?(insn) && insn != :opt_regexpmatch2
       $stderr.puts
       warn "'#{insn}' insn is not included in the script. Actual insns are: #{used_insns.join(' ')}\n", uplevel: uplevel
     end
