@@ -8,7 +8,7 @@ static VALUE rb_eClosedQueueError;
 typedef struct rb_mutex_struct {
     rb_fiber_t *fiber;
     struct rb_mutex_struct *next_mutex;
-    struct list_head waitq; /* protected by GVL */
+    struct ccan_list_head waitq; /* protected by GVL */
 } rb_mutex_t;
 
 /* sync_waiter is always on-stack */
@@ -16,41 +16,42 @@ struct sync_waiter {
     VALUE self;
     rb_thread_t *th;
     rb_fiber_t *fiber;
-    struct list_node node;
+    struct ccan_list_node node;
 };
 
 #define MUTEX_ALLOW_TRAP FL_USER1
 
 static void
-sync_wakeup(struct list_head *head, long max)
+sync_wakeup(struct ccan_list_head *head, long max)
 {
     struct sync_waiter *cur = 0, *next;
 
-    list_for_each_safe(head, cur, next, node) {
-        list_del_init(&cur->node);
-
-        if (cur->th->scheduler != Qnil) {
-            rb_scheduler_unblock(cur->th->scheduler, cur->self, rb_fiberptr_self(cur->fiber));
-        }
+    ccan_list_for_each_safe(head, cur, next, node) {
+        ccan_list_del_init(&cur->node);
 
         if (cur->th->status != THREAD_KILLED) {
-            if (cur->th->scheduler == Qnil) {
+
+            if (cur->th->scheduler != Qnil && rb_fiberptr_blocking(cur->fiber) == 0) {
+                rb_fiber_scheduler_unblock(cur->th->scheduler, cur->self, rb_fiberptr_self(cur->fiber));
+            }
+            else {
                 rb_threadptr_interrupt(cur->th);
                 cur->th->status = THREAD_RUNNABLE;
             }
+
             if (--max == 0) return;
         }
     }
 }
 
 static void
-wakeup_one(struct list_head *head)
+wakeup_one(struct ccan_list_head *head)
 {
     sync_wakeup(head, 1);
 }
 
 static void
-wakeup_all(struct list_head *head)
+wakeup_all(struct ccan_list_head *head)
 {
     sync_wakeup(head, LONG_MAX);
 }
@@ -63,14 +64,14 @@ static void rb_mutex_abandon_locking_mutex(rb_thread_t *th);
 static const char* rb_mutex_unlock_th(rb_mutex_t *mutex, rb_thread_t *th, rb_fiber_t *fiber);
 
 /*
- *  Document-class: Mutex
+ *  Document-class: Thread::Mutex
  *
- *  Mutex implements a simple semaphore that can be used to coordinate access to
- *  shared data from multiple concurrent threads.
+ *  Thread::Mutex implements a simple semaphore that can be used to
+ *  coordinate access to shared data from multiple concurrent threads.
  *
  *  Example:
  *
- *    semaphore = Mutex.new
+ *    semaphore = Thread::Mutex.new
  *
  *    a = Thread.new {
  *      semaphore.synchronize {
@@ -94,7 +95,7 @@ rb_mutex_num_waiting(rb_mutex_t *mutex)
     struct sync_waiter *w = 0;
     size_t n = 0;
 
-    list_for_each(&mutex->waitq, w, node) {
+    ccan_list_for_each(&mutex->waitq, w, node) {
 	n++;
     }
 
@@ -124,7 +125,7 @@ mutex_memsize(const void *ptr)
 static const rb_data_type_t mutex_data_type = {
     "mutex",
     {mutex_mark, mutex_free, mutex_memsize,},
-    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
+    0, 0, RUBY_TYPED_WB_PROTECTED | RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 static rb_mutex_t *
@@ -140,12 +141,7 @@ mutex_ptr(VALUE obj)
 VALUE
 rb_obj_is_mutex(VALUE obj)
 {
-    if (rb_typeddata_is_kind_of(obj, &mutex_data_type)) {
-	return Qtrue;
-    }
-    else {
-	return Qfalse;
-    }
+    return RBOOL(rb_typeddata_is_kind_of(obj, &mutex_data_type));
 }
 
 static VALUE
@@ -156,13 +152,13 @@ mutex_alloc(VALUE klass)
 
     obj = TypedData_Make_Struct(klass, rb_mutex_t, &mutex_data_type, mutex);
 
-    list_head_init(&mutex->waitq);
+    ccan_list_head_init(&mutex->waitq);
     return obj;
 }
 
 /*
  *  call-seq:
- *     Mutex.new   -> mutex
+ *     Thread::Mutex.new   -> mutex
  *
  *  Creates a new Mutex
  */
@@ -189,7 +185,33 @@ rb_mutex_locked_p(VALUE self)
 {
     rb_mutex_t *mutex = mutex_ptr(self);
 
-    return mutex->fiber ? Qtrue : Qfalse;
+    return RBOOL(mutex->fiber);
+}
+
+static void
+thread_mutex_insert(rb_thread_t *thread, rb_mutex_t *mutex)
+{
+    if (thread->keeping_mutexes) {
+        mutex->next_mutex = thread->keeping_mutexes;
+    }
+
+    thread->keeping_mutexes = mutex;
+}
+
+static void
+thread_mutex_remove(rb_thread_t *thread, rb_mutex_t *mutex)
+{
+    rb_mutex_t **keeping_mutexes = &thread->keeping_mutexes;
+
+    while (*keeping_mutexes && *keeping_mutexes != mutex) {
+        // Move to the next mutex in the list:
+        keeping_mutexes = &(*keeping_mutexes)->next_mutex;
+    }
+
+    if (*keeping_mutexes) {
+        *keeping_mutexes = mutex->next_mutex;
+        mutex->next_mutex = NULL;
+    }
 }
 
 static void
@@ -197,10 +219,7 @@ mutex_locked(rb_thread_t *th, VALUE self)
 {
     rb_mutex_t *mutex = mutex_ptr(self);
 
-    if (th->keeping_mutexes) {
-	mutex->next_mutex = th->keeping_mutexes;
-    }
-    th->keeping_mutexes = mutex;
+    thread_mutex_insert(th, mutex);
 }
 
 /*
@@ -237,21 +256,21 @@ static const rb_thread_t *patrol_thread = NULL;
 static VALUE
 mutex_owned_p(rb_fiber_t *fiber, rb_mutex_t *mutex)
 {
-    if (mutex->fiber == fiber) {
-        return Qtrue;
-    }
-    else {
-        return Qfalse;
-    }
+    return RBOOL(mutex->fiber == fiber);
 }
 
-static VALUE call_rb_scheduler_block(VALUE mutex) {
-    return rb_scheduler_block(rb_thread_current_scheduler(), mutex, Qnil);
+static VALUE
+call_rb_fiber_scheduler_block(VALUE mutex)
+{
+    return rb_fiber_scheduler_block(rb_fiber_scheduler_current(), mutex, Qnil);
 }
 
-static VALUE remove_from_mutex_lock_waiters(VALUE arg) {
-    struct list_node *node = (struct list_node*)arg;
-    list_del(node);
+static VALUE
+delete_from_waitq(VALUE value)
+{
+    struct sync_waiter *sync_waiter = (void *)value;
+    ccan_list_del(&sync_waiter->node);
+
     return Qnil;
 }
 
@@ -270,27 +289,28 @@ do_mutex_lock(VALUE self, int interruptible_p)
     }
 
     if (rb_mutex_trylock(self) == Qfalse) {
-        struct sync_waiter w = {
-            .self = self,
-            .th = th,
-            .fiber = fiber
-        };
-
         if (mutex->fiber == fiber) {
             rb_raise(rb_eThreadError, "deadlock; recursive locking");
         }
 
         while (mutex->fiber != fiber) {
-            VALUE scheduler = rb_thread_current_scheduler();
+            VALUE scheduler = rb_fiber_scheduler_current();
             if (scheduler != Qnil) {
-                list_add_tail(&mutex->waitq, &w.node);
+                struct sync_waiter sync_waiter = {
+                    .self = self,
+                    .th = th,
+                    .fiber = fiber
+                };
 
-                rb_ensure(call_rb_scheduler_block, self, remove_from_mutex_lock_waiters, (VALUE)&w.node);
+                ccan_list_add_tail(&mutex->waitq, &sync_waiter.node);
+
+                rb_ensure(call_rb_fiber_scheduler_block, self, delete_from_waitq, (VALUE)&sync_waiter);
 
                 if (!mutex->fiber) {
                     mutex->fiber = fiber;
                 }
-            } else {
+            }
+            else {
                 enum rb_thread_status prev_status = th->status;
                 rb_hrtime_t *timeout = 0;
                 rb_hrtime_t rel = rb_msec2hrtime(100);
@@ -309,11 +329,17 @@ do_mutex_lock(VALUE self, int interruptible_p)
                     patrol_thread = th;
                 }
 
-                list_add_tail(&mutex->waitq, &w.node);
+                struct sync_waiter sync_waiter = {
+                    .self = self,
+                    .th = th,
+                    .fiber = fiber
+                };
+
+                ccan_list_add_tail(&mutex->waitq, &sync_waiter.node);
 
                 native_sleep(th, timeout); /* release GVL */
 
-                list_del(&w.node);
+                ccan_list_del(&sync_waiter.node);
 
                 if (!mutex->fiber) {
                     mutex->fiber = fiber;
@@ -392,23 +418,23 @@ rb_mutex_unlock_th(rb_mutex_t *mutex, rb_thread_t *th, rb_fiber_t *fiber)
     const char *err = NULL;
 
     if (mutex->fiber == 0) {
-	err = "Attempt to unlock a mutex which is not locked";
+        err = "Attempt to unlock a mutex which is not locked";
     }
     else if (mutex->fiber != fiber) {
-	err = "Attempt to unlock a mutex which is locked by another thread/fiber";
+        err = "Attempt to unlock a mutex which is locked by another thread/fiber";
     }
     else {
-	struct sync_waiter *cur = 0, *next;
-	rb_mutex_t **th_mutex = &th->keeping_mutexes;
+        struct sync_waiter *cur = 0, *next;
 
-	mutex->fiber = 0;
-	list_for_each_safe(&mutex->waitq, cur, next, node) {
-	    list_del_init(&cur->node);
+        mutex->fiber = 0;
+        ccan_list_for_each_safe(&mutex->waitq, cur, next, node) {
+            ccan_list_del_init(&cur->node);
 
-            if (cur->th->scheduler != Qnil) {
-                rb_scheduler_unblock(cur->th->scheduler, cur->self, rb_fiberptr_self(cur->fiber));
+            if (cur->th->scheduler != Qnil && rb_fiberptr_blocking(cur->fiber) == 0) {
+                rb_fiber_scheduler_unblock(cur->th->scheduler, cur->self, rb_fiberptr_self(cur->fiber));
                 goto found;
-            } else {
+            }
+            else {
                 switch (cur->th->status) {
                   case THREAD_RUNNABLE: /* from someone else calling Thread#run */
                   case THREAD_STOPPED_FOREVER: /* likely (rb_mutex_lock) */
@@ -422,13 +448,10 @@ rb_mutex_unlock_th(rb_mutex_t *mutex, rb_thread_t *th, rb_fiber_t *fiber)
                     continue;
                 }
             }
-	}
-      found:
-	while (*th_mutex != mutex) {
-	    th_mutex = &(*th_mutex)->next_mutex;
-	}
-	*th_mutex = mutex->next_mutex;
-	mutex->next_mutex = NULL;
+        }
+
+    found:
+        thread_mutex_remove(th, mutex);
     }
 
     return err;
@@ -468,7 +491,7 @@ rb_mutex_abandon_locking_mutex(rb_thread_t *th)
     if (th->locking_mutex) {
         rb_mutex_t *mutex = mutex_ptr(th->locking_mutex);
 
-        list_head_init(&mutex->waitq);
+        ccan_list_head_init(&mutex->waitq);
         th->locking_mutex = Qfalse;
     }
 }
@@ -483,7 +506,7 @@ rb_mutex_abandon_all(rb_mutex_t *mutexes)
 	mutexes = mutex->next_mutex;
 	mutex->fiber = 0;
 	mutex->next_mutex = 0;
-	list_head_init(&mutex->waitq);
+	ccan_list_head_init(&mutex->waitq);
     }
 }
 #endif
@@ -500,14 +523,14 @@ rb_mutex_wait_for(VALUE time)
 {
     rb_hrtime_t *rel = (rb_hrtime_t *)time;
     /* permit spurious check */
-    sleep_hrtime(GET_THREAD(), *rel, 0);
-    return Qnil;
+    return RBOOL(sleep_hrtime(GET_THREAD(), *rel, 0));
 }
 
 VALUE
 rb_mutex_sleep(VALUE self, VALUE timeout)
 {
     struct timeval t;
+    VALUE woken = Qtrue;
 
     if (!NIL_P(timeout)) {
         t = rb_time_interval(timeout);
@@ -516,27 +539,30 @@ rb_mutex_sleep(VALUE self, VALUE timeout)
     rb_mutex_unlock(self);
     time_t beg = time(0);
 
-    VALUE scheduler = rb_thread_current_scheduler();
+    VALUE scheduler = rb_fiber_scheduler_current();
     if (scheduler != Qnil) {
-        rb_scheduler_kernel_sleep(scheduler, timeout);
+        rb_fiber_scheduler_kernel_sleep(scheduler, timeout);
         mutex_lock_uninterruptible(self);
-    } else {
+    }
+    else {
         if (NIL_P(timeout)) {
             rb_ensure(rb_mutex_sleep_forever, self, mutex_lock_uninterruptible, self);
-        } else {
+        }
+        else {
             rb_hrtime_t rel = rb_timeval2hrtime(&t);
-            rb_ensure(rb_mutex_wait_for, (VALUE)&rel, mutex_lock_uninterruptible, self);
+            woken = rb_ensure(rb_mutex_wait_for, (VALUE)&rel, mutex_lock_uninterruptible, self);
         }
     }
 
     RUBY_VM_CHECK_INTS_BLOCKING(GET_EC());
+    if (!woken) return Qnil;
     time_t end = time(0) - beg;
-    return INT2FIX(end);
+    return TIMET2NUM(end);
 }
 
 /*
  * call-seq:
- *    mutex.sleep(timeout = nil)    -> number
+ *    mutex.sleep(timeout = nil)    -> number or nil
  *
  * Releases the lock and sleeps +timeout+ seconds if it is given and
  * non-nil or forever.  Raises +ThreadError+ if +mutex+ wasn't locked by
@@ -547,6 +573,8 @@ rb_mutex_sleep(VALUE self, VALUE timeout)
  *
  * Note that this method can wakeup without explicit Thread#wakeup call.
  * For example, receiving signal and so on.
+ *
+ * Returns the slept time in seconds if woken up, or +nil+ if timed out.
  */
 static VALUE
 mutex_sleep(int argc, VALUE *argv, VALUE self)
@@ -562,7 +590,7 @@ mutex_sleep(int argc, VALUE *argv, VALUE self)
  *    mutex.synchronize { ... }    -> result of the block
  *
  * Obtains a lock, runs the block, and releases the lock when the block
- * completes.  See the example under +Mutex+.
+ * completes.  See the example under Thread::Mutex.
  */
 
 VALUE
@@ -577,7 +605,7 @@ rb_mutex_synchronize(VALUE mutex, VALUE (*func)(VALUE arg), VALUE arg)
  *    mutex.synchronize { ... }    -> result of the block
  *
  * Obtains a lock, runs the block, and releases the lock when the block
- * completes.  See the example under +Mutex+.
+ * completes.  See the example under Thread::Mutex.
  */
 static VALUE
 rb_mutex_synchronize_m(VALUE self)
@@ -603,7 +631,7 @@ void rb_mutex_allow_trap(VALUE self, int val)
 
 #define queue_waitq(q) UNALIGNED_MEMBER_PTR(q, waitq)
 PACKED_STRUCT_UNALIGNED(struct rb_queue {
-    struct list_head waitq;
+    struct ccan_list_head waitq;
     rb_serial_t fork_gen;
     const VALUE que;
     int num_waiting;
@@ -614,7 +642,7 @@ PACKED_STRUCT_UNALIGNED(struct rb_queue {
 PACKED_STRUCT_UNALIGNED(struct rb_szqueue {
     struct rb_queue q;
     int num_waiting_push;
-    struct list_head pushq;
+    struct ccan_list_head pushq;
     long max;
 });
 
@@ -646,7 +674,7 @@ queue_alloc(VALUE klass)
     struct rb_queue *q;
 
     obj = TypedData_Make_Struct(klass, struct rb_queue, &queue_data_type, q);
-    list_head_init(queue_waitq(q));
+    ccan_list_head_init(queue_waitq(q));
     return obj;
 }
 
@@ -660,7 +688,7 @@ queue_fork_check(struct rb_queue *q)
     }
     /* forked children can't reach into parent thread stacks */
     q->fork_gen = fork_gen;
-    list_head_init(queue_waitq(q));
+    ccan_list_head_init(queue_waitq(q));
     q->num_waiting = 0;
     return 1;
 }
@@ -704,8 +732,8 @@ szqueue_alloc(VALUE klass)
     struct rb_szqueue *sq;
     VALUE obj = TypedData_Make_Struct(klass, struct rb_szqueue,
 					&szqueue_data_type, sq);
-    list_head_init(szqueue_waitq(sq));
-    list_head_init(szqueue_pushq(sq));
+    ccan_list_head_init(szqueue_waitq(sq));
+    ccan_list_head_init(szqueue_pushq(sq));
     return obj;
 }
 
@@ -716,7 +744,7 @@ szqueue_ptr(VALUE obj)
 
     TypedData_Get_Struct(obj, struct rb_szqueue, &szqueue_data_type, sq);
     if (queue_fork_check(&sq->q)) {
-        list_head_init(szqueue_pushq(sq));
+        ccan_list_head_init(szqueue_pushq(sq));
         sq->num_waiting_push = 0;
     }
 
@@ -754,7 +782,7 @@ queue_closed_p(VALUE self)
  *  Document-class: ClosedQueueError
  *
  *  The exception class which will be raised when pushing into a closed
- *  Queue.  See Queue#close and SizedQueue#close.
+ *  Queue.  See Thread::Queue#close and Thread::SizedQueue#close.
  */
 
 NORETURN(static void raise_closed_queue_error(VALUE self));
@@ -773,19 +801,19 @@ queue_closed_result(VALUE self, struct rb_queue *q)
 }
 
 /*
- *  Document-class: Queue
+ *  Document-class: Thread::Queue
  *
- *  The Queue class implements multi-producer, multi-consumer queues.
- *  It is especially useful in threaded programming when information
- *  must be exchanged safely between multiple threads. The Queue class
- *  implements all the required locking semantics.
+ *  The Thread::Queue class implements multi-producer, multi-consumer
+ *  queues.  It is especially useful in threaded programming when
+ *  information must be exchanged safely between multiple threads. The
+ *  Thread::Queue class implements all the required locking semantics.
  *
  *  The class implements FIFO type of queue. In a FIFO queue, the first
  *  tasks added are the first retrieved.
  *
  *  Example:
  *
- *	queue = Queue.new
+ *	queue = Thread::Queue.new
  *
  *	producer = Thread.new do
  *	  5.times do |i|
@@ -810,15 +838,41 @@ queue_closed_result(VALUE self, struct rb_queue *q)
 /*
  * Document-method: Queue::new
  *
- * Creates a new queue instance.
+ * call-seq:
+ *   Thread::Queue.new -> empty_queue
+ *   Thread::Queue.new(enumerable) -> queue
+ *
+ * Creates a new queue instance, optionally using the contents of an +enumerable+
+ * for its initial state.
+ *
+ * Example:
+ *
+ *    	q = Thread::Queue.new
+ *      #=> #<Thread::Queue:0x00007ff7501110d0>
+ *      q.empty?
+ *      #=> true
+ *
+ *    	q = Thread::Queue.new([1, 2, 3])
+ *    	#=> #<Thread::Queue:0x00007ff7500ec500>
+ *      q.empty?
+ *      #=> false
+ *      q.pop
+ *      #=> 1
  */
 
 static VALUE
-rb_queue_initialize(VALUE self)
+rb_queue_initialize(int argc, VALUE *argv, VALUE self)
 {
+    VALUE initial;
     struct rb_queue *q = queue_ptr(self);
+    if ((argc = rb_scan_args(argc, argv, "01", &initial)) == 1) {
+        initial = rb_to_array(initial);
+    }
     RB_OBJ_WRITE(self, &q->que, ary_buf_new());
-    list_head_init(queue_waitq(q));
+    ccan_list_head_init(queue_waitq(q));
+    if (argc == 1) {
+        rb_ary_concat(q->que, initial);
+    }
     return self;
 }
 
@@ -834,7 +888,7 @@ queue_do_push(VALUE self, struct rb_queue *q, VALUE obj)
 }
 
 /*
- * Document-method: Queue#close
+ * Document-method: Thread::Queue#close
  * call-seq:
  *   close
  *
@@ -855,9 +909,9 @@ queue_do_push(VALUE self, struct rb_queue *q, VALUE obj)
  *
  * ClosedQueueError is inherited from StopIteration, so that you can break loop block.
  *
- *  Example:
+ * Example:
  *
- *    	q = Queue.new
+ *    	q = Thread::Queue.new
  *      Thread.new{
  *        while e = q.deq # wait for nil to break loop
  *          # ...
@@ -881,7 +935,7 @@ rb_queue_close(VALUE self)
 }
 
 /*
- * Document-method: Queue#closed?
+ * Document-method: Thread::Queue#closed?
  * call-seq: closed?
  *
  * Returns +true+ if the queue is closed.
@@ -890,11 +944,11 @@ rb_queue_close(VALUE self)
 static VALUE
 rb_queue_closed_p(VALUE self)
 {
-    return queue_closed_p(self) ? Qtrue : Qfalse;
+    return RBOOL(queue_closed_p(self));
 }
 
 /*
- * Document-method: Queue#push
+ * Document-method: Thread::Queue#push
  * call-seq:
  *   push(object)
  *   enq(object)
@@ -929,7 +983,7 @@ queue_sleep_done(VALUE p)
 {
     struct queue_waiter *qw = (struct queue_waiter *)p;
 
-    list_del(&qw->w.node);
+    ccan_list_del(&qw->w.node);
     qw->as.q->num_waiting--;
 
     return Qfalse;
@@ -940,7 +994,7 @@ szqueue_sleep_done(VALUE p)
 {
     struct queue_waiter *qw = (struct queue_waiter *)p;
 
-    list_del(&qw->w.node);
+    ccan_list_del(&qw->w.node);
     qw->as.sq->num_waiting_push--;
 
     return Qfalse;
@@ -960,20 +1014,21 @@ queue_do_pop(VALUE self, struct rb_queue *q, int should_block)
         }
         else {
             rb_execution_context_t *ec = GET_EC();
-            struct queue_waiter qw;
 
             assert(RARRAY_LEN(q->que) == 0);
             assert(queue_closed_p(self) == 0);
 
-            qw.w.self = self;
-            qw.w.th = ec->thread_ptr;
-            qw.w.fiber = ec->fiber_ptr;
+            struct queue_waiter queue_waiter = {
+                .w = {.self = self, .th = ec->thread_ptr, .fiber = ec->fiber_ptr},
+                .as = {.q = q}
+            };
 
-            qw.as.q = q;
-            list_add_tail(queue_waitq(qw.as.q), &qw.w.node);
-            qw.as.q->num_waiting++;
+            struct ccan_list_head *waitq = queue_waitq(q);
 
-            rb_ensure(queue_sleep, self, queue_sleep_done, (VALUE)&qw);
+            ccan_list_add_tail(waitq, &queue_waiter.w.node);
+            queue_waiter.as.q->num_waiting++;
+
+            rb_ensure(queue_sleep, self, queue_sleep_done, (VALUE)&queue_waiter);
         }
     }
 
@@ -992,7 +1047,7 @@ queue_pop_should_block(int argc, const VALUE *argv)
 }
 
 /*
- * Document-method: Queue#pop
+ * Document-method: Thread::Queue#pop
  * call-seq:
  *   pop(non_block=false)
  *   deq(non_block=false)
@@ -1013,7 +1068,7 @@ rb_queue_pop(int argc, VALUE *argv, VALUE self)
 }
 
 /*
- * Document-method: Queue#empty?
+ * Document-method: Thread::Queue#empty?
  * call-seq: empty?
  *
  * Returns +true+ if the queue is empty.
@@ -1022,11 +1077,11 @@ rb_queue_pop(int argc, VALUE *argv, VALUE self)
 static VALUE
 rb_queue_empty_p(VALUE self)
 {
-    return queue_length(self, queue_ptr(self)) == 0 ? Qtrue : Qfalse;
+    return RBOOL(queue_length(self, queue_ptr(self)) == 0);
 }
 
 /*
- * Document-method: Queue#clear
+ * Document-method: Thread::Queue#clear
  *
  * Removes all objects from the queue.
  */
@@ -1041,7 +1096,7 @@ rb_queue_clear(VALUE self)
 }
 
 /*
- * Document-method: Queue#length
+ * Document-method: Thread::Queue#length
  * call-seq:
  *   length
  *   size
@@ -1056,7 +1111,7 @@ rb_queue_length(VALUE self)
 }
 
 /*
- * Document-method: Queue#num_waiting
+ * Document-method: Thread::Queue#num_waiting
  *
  * Returns the number of threads waiting on the queue.
  */
@@ -1070,12 +1125,12 @@ rb_queue_num_waiting(VALUE self)
 }
 
 /*
- *  Document-class: SizedQueue
+ *  Document-class: Thread::SizedQueue
  *
  * This class represents queues of specified size capacity.  The push operation
  * may be blocked if the capacity is full.
  *
- * See Queue for an example of how a SizedQueue works.
+ * See Thread::Queue for an example of how a Thread::SizedQueue works.
  */
 
 /*
@@ -1097,19 +1152,19 @@ rb_szqueue_initialize(VALUE self, VALUE vmax)
     }
 
     RB_OBJ_WRITE(self, &sq->q.que, ary_buf_new());
-    list_head_init(szqueue_waitq(sq));
-    list_head_init(szqueue_pushq(sq));
+    ccan_list_head_init(szqueue_waitq(sq));
+    ccan_list_head_init(szqueue_pushq(sq));
     sq->max = max;
 
     return self;
 }
 
 /*
- * Document-method: SizedQueue#close
+ * Document-method: Thread::SizedQueue#close
  * call-seq:
  *   close
  *
- * Similar to Queue#close.
+ * Similar to Thread::Queue#close.
  *
  * The difference is behavior with waiting enqueuing threads.
  *
@@ -1130,7 +1185,7 @@ rb_szqueue_close(VALUE self)
 }
 
 /*
- * Document-method: SizedQueue#max
+ * Document-method: Thread::SizedQueue#max
  *
  * Returns the maximum size of the queue.
  */
@@ -1142,7 +1197,7 @@ rb_szqueue_max_get(VALUE self)
 }
 
 /*
- * Document-method: SizedQueue#max=
+ * Document-method: Thread::SizedQueue#max=
  * call-seq: max=(number)
  *
  * Sets the maximum size of the queue to the given +number+.
@@ -1178,7 +1233,7 @@ szqueue_push_should_block(int argc, const VALUE *argv)
 }
 
 /*
- * Document-method: SizedQueue#push
+ * Document-method: Thread::SizedQueue#push
  * call-seq:
  *   push(object, non_block=false)
  *   enq(object, non_block=false)
@@ -1206,18 +1261,17 @@ rb_szqueue_push(int argc, VALUE *argv, VALUE self)
         }
         else {
             rb_execution_context_t *ec = GET_EC();
-            struct queue_waiter qw;
-            struct list_head *pushq = szqueue_pushq(sq);
+            struct queue_waiter queue_waiter = {
+                .w = {.self = self, .th = ec->thread_ptr, .fiber = ec->fiber_ptr},
+                .as = {.sq = sq}
+            };
 
-            qw.w.self = self;
-            qw.w.th = ec->thread_ptr;
-            qw.w.fiber = ec->fiber_ptr;
+            struct ccan_list_head *pushq = szqueue_pushq(sq);
 
-            qw.as.sq = sq;
-            list_add_tail(pushq, &qw.w.node);
+            ccan_list_add_tail(pushq, &queue_waiter.w.node);
             sq->num_waiting_push++;
 
-            rb_ensure(queue_sleep, self, szqueue_sleep_done, (VALUE)&qw);
+            rb_ensure(queue_sleep, self, szqueue_sleep_done, (VALUE)&queue_waiter);
         }
     }
 
@@ -1242,7 +1296,7 @@ szqueue_do_pop(VALUE self, int should_block)
 }
 
 /*
- * Document-method: SizedQueue#pop
+ * Document-method: Thread::SizedQueue#pop
  * call-seq:
  *   pop(non_block=false)
  *   deq(non_block=false)
@@ -1263,7 +1317,7 @@ rb_szqueue_pop(int argc, VALUE *argv, VALUE self)
 }
 
 /*
- * Document-method: SizedQueue#clear
+ * Document-method: Thread::SizedQueue#clear
  *
  * Removes all objects from the queue.
  */
@@ -1279,7 +1333,7 @@ rb_szqueue_clear(VALUE self)
 }
 
 /*
- * Document-method: SizedQueue#length
+ * Document-method: Thread::SizedQueue#length
  * call-seq:
  *   length
  *   size
@@ -1296,7 +1350,7 @@ rb_szqueue_length(VALUE self)
 }
 
 /*
- * Document-method: SizedQueue#num_waiting
+ * Document-method: Thread::SizedQueue#num_waiting
  *
  * Returns the number of threads waiting on the queue.
  */
@@ -1310,7 +1364,7 @@ rb_szqueue_num_waiting(VALUE self)
 }
 
 /*
- * Document-method: SizedQueue#empty?
+ * Document-method: Thread::SizedQueue#empty?
  * call-seq: empty?
  *
  * Returns +true+ if the queue is empty.
@@ -1321,18 +1375,18 @@ rb_szqueue_empty_p(VALUE self)
 {
     struct rb_szqueue *sq = szqueue_ptr(self);
 
-    return queue_length(self, &sq->q) == 0 ? Qtrue : Qfalse;
+    return RBOOL(queue_length(self, &sq->q) == 0);
 }
 
 
 /* ConditionalVariable */
 struct rb_condvar {
-    struct list_head waitq;
+    struct ccan_list_head waitq;
     rb_serial_t fork_gen;
 };
 
 /*
- *  Document-class: ConditionVariable
+ *  Document-class: Thread::ConditionVariable
  *
  *  ConditionVariable objects augment class Mutex. Using condition variables,
  *  it is possible to suspend while in the middle of a critical section until a
@@ -1340,8 +1394,8 @@ struct rb_condvar {
  *
  *  Example:
  *
- *    mutex = Mutex.new
- *    resource = ConditionVariable.new
+ *    mutex = Thread::Mutex.new
+ *    resource = Thread::ConditionVariable.new
  *
  *    a = Thread.new {
  *	 mutex.synchronize {
@@ -1382,7 +1436,7 @@ condvar_ptr(VALUE self)
     /* forked children can't reach into parent thread stacks */
     if (cv->fork_gen != fork_gen) {
         cv->fork_gen = fork_gen;
-        list_head_init(&cv->waitq);
+        ccan_list_head_init(&cv->waitq);
     }
 
     return cv;
@@ -1395,7 +1449,7 @@ condvar_alloc(VALUE klass)
     VALUE obj;
 
     obj = TypedData_Make_Struct(klass, struct rb_condvar, &cv_data_type, cv);
-    list_head_init(&cv->waitq);
+    ccan_list_head_init(&cv->waitq);
 
     return obj;
 }
@@ -1410,7 +1464,7 @@ static VALUE
 rb_condvar_initialize(VALUE self)
 {
     struct rb_condvar *cv = condvar_ptr(self);
-    list_head_init(&cv->waitq);
+    ccan_list_head_init(&cv->waitq);
     return self;
 }
 
@@ -1428,23 +1482,16 @@ do_sleep(VALUE args)
     return rb_funcallv(p->mutex, id_sleep, 1, &p->timeout);
 }
 
-static VALUE
-delete_from_waitq(VALUE v)
-{
-    struct sync_waiter *w = (void *)v;
-    list_del(&w->node);
-
-    return Qnil;
-}
-
 /*
- * Document-method: ConditionVariable#wait
+ * Document-method: Thread::ConditionVariable#wait
  * call-seq: wait(mutex, timeout=nil)
  *
  * Releases the lock held in +mutex+ and waits; reacquires the lock on wakeup.
  *
  * If +timeout+ is given, this method returns after +timeout+ seconds passed,
  * even if no other thread doesn't signal.
+ *
+ * Returns the slept result on +mutex+.
  */
 
 static VALUE
@@ -1457,20 +1504,18 @@ rb_condvar_wait(int argc, VALUE *argv, VALUE self)
 
     rb_scan_args(argc, argv, "11", &args.mutex, &args.timeout);
 
-    struct sync_waiter w = {
+    struct sync_waiter sync_waiter = {
         .self = args.mutex,
         .th = ec->thread_ptr,
-        .fiber = ec->fiber_ptr,
+        .fiber = ec->fiber_ptr
     };
 
-    list_add_tail(&cv->waitq, &w.node);
-    rb_ensure(do_sleep, (VALUE)&args, delete_from_waitq, (VALUE)&w);
-
-    return self;
+    ccan_list_add_tail(&cv->waitq, &sync_waiter.node);
+    return rb_ensure(do_sleep, (VALUE)&args, delete_from_waitq, (VALUE)&sync_waiter);
 }
 
 /*
- * Document-method: ConditionVariable#signal
+ * Document-method: Thread::ConditionVariable#signal
  *
  * Wakes up the first thread in line waiting for this lock.
  */
@@ -1484,7 +1529,7 @@ rb_condvar_signal(VALUE self)
 }
 
 /*
- * Document-method: ConditionVariable#broadcast
+ * Document-method: Thread::ConditionVariable#broadcast
  *
  * Wakes up all threads waiting for this lock.
  */
@@ -1507,10 +1552,10 @@ undumpable(VALUE obj)
 }
 
 static VALUE
-define_thread_class(VALUE outer, const char *name, VALUE super)
+define_thread_class(VALUE outer, const ID name, VALUE super)
 {
-    VALUE klass = rb_define_class_under(outer, name, super);
-    rb_define_const(rb_cObject, name, klass);
+    VALUE klass = rb_define_class_id_under(outer, name, super);
+    rb_const_set(rb_cObject, name, klass);
     return klass;
 }
 
@@ -1518,15 +1563,15 @@ static void
 Init_thread_sync(void)
 {
 #undef rb_intern
-#if 0
-    rb_cMutex = rb_define_class("Mutex", rb_cObject); /* teach rdoc Mutex */
-    rb_cConditionVariable = rb_define_class("ConditionVariable", rb_cObject); /* teach rdoc ConditionVariable */
-    rb_cQueue = rb_define_class("Queue", rb_cObject); /* teach rdoc Queue */
-    rb_cSizedQueue = rb_define_class("SizedQueue", rb_cObject); /* teach rdoc SizedQueue */
+#if defined(TEACH_RDOC) && TEACH_RDOC == 42
+    rb_cMutex = rb_define_class_under(rb_cThread, "Mutex", rb_cObject);
+    rb_cConditionVariable = rb_define_class_under(rb_cThread, "ConditionVariable", rb_cObject);
+    rb_cQueue = rb_define_class_under(rb_cThread, "Queue", rb_cObject);
+    rb_cSizedQueue = rb_define_class_under(rb_cThread, "SizedQueue", rb_cObject);
 #endif
 
 #define DEFINE_CLASS(name, super) \
-    rb_c##name = define_thread_class(rb_cThread, #name, rb_c##super)
+    rb_c##name = define_thread_class(rb_cThread, rb_intern(#name), rb_c##super)
 
     /* Mutex */
     DEFINE_CLASS(Mutex, Object);
@@ -1546,7 +1591,7 @@ Init_thread_sync(void)
 
     rb_eClosedQueueError = rb_define_class("ClosedQueueError", rb_eStopIteration);
 
-    rb_define_method(rb_cQueue, "initialize", rb_queue_initialize, 0);
+    rb_define_method(rb_cQueue, "initialize", rb_queue_initialize, -1);
     rb_undef_method(rb_cQueue, "initialize_copy");
     rb_define_method(rb_cQueue, "marshal_dump", undumpable, 0);
     rb_define_method(rb_cQueue, "close", rb_queue_close, 0);
